/*
 *
 * graph. c                  -- graph relation procedures
 *
 * Copyright ??2002 Sadao Sato
 *
 *
 * Permission to use, copy, modify, distribute,and license this
 * software and its documentation for any purpose is hereby granted,
 * provided that existing copyright notices are retained in all
 * copies and that this notice is included verbatim in any
 * distributions.  No written agreement, license, or royalty fee is
 * required for any of the authorized uses.
 * This software is provided ``AS IS'' without express or implied
 * warranty.
 *
 *
 *           Author: Sadao Sato
 *    Creation date: 13-Dec-2002
 * Last file update: 10-June-2004
 */

/***  conn-list is a dot-list of graph edge (no oriented)
***/  

#include "lisp.h"

ptr connectp();

ptr Err(char *p,int errnum, ptr x){
 /*** printf("%s \n",p);***/
 lwritestr(STDERR,p);
 newline(STDERR);
 return error(errnum,x);
}

ptr s_co_vertex(int narg, ptr env)
{
  ptr obj,  dot;
  checknarg(narg-2);
   obj=*(stack-2);
   dot=*(stack-1);
  if (!CONSP(dot))  Err("co-pair: not a dot-pair", E_notlist,dot);
   if(equal(CAR(dot), obj) == Truth) return CDR(dot);
   if(equal(CDR(dot), obj) == Truth) return CAR(dot);
   return Ntruth;
}

ptr s_find_edge(int narg, ptr env)
{
  Reg  ptr obj,  elist;
  Reg  ptr l,tmp;
  checknarg(narg-2);
   obj=*(stack-2);
   elist=*(stack-1);

  for(l=elist; CONSP(l); ) {
    tmp = CAR(l);
    if (CONSP(tmp)){
      if(equal(CAR(tmp), obj) == Truth) return tmp;
      if(equal(CDR(tmp), obj) == Truth) return tmp;
    }
    if ((l=CDR(l)) == elist) goto Error;
  }
  if (NULLP(l)) return(Ntruth);
Error:
  Err("connection function: improper list", E_notlist,elist);
  return UNDEFINED; /* never reached */
}


ptr s_nearest(int narg, ptr  env)
{
   Reg ptr obj,  elist;
  Reg  ptr l,x,tmp,z;
  checknarg(narg-2);
   obj=*(stack-2);
   elist=*(stack-1);

  tmp=z=CONS(NIL,NIL);
  psh(z);
  for(l=elist; CONSP(l); ) {
    x = CAR(l);
    if (CONSP(x)){
      if(equal(CAR(x), obj) == Truth)
        tmp=CDR(tmp)=CONS(CDR(x),NIL);
      if(equal(CDR(x), obj) == Truth)
        tmp=CDR(tmp)=CONS(CAR(x),NIL);
    }
    if ((l=CDR(l)) == elist) goto Error;
  }
  pull();
   return CDR(z);
Error:
  pull();
  Err("connection function: improper list", E_notlist,elist);
  return UNDEFINED; /* never reached */
}

ptr s_connectp(int narg, ptr env)
{
  Reg ptr obj1, obj2, elist;
   checknarg(narg-3);
   obj1=*(stack-3);
   obj2=*(stack-2);
   elist=*(stack-1);
   return connectp(obj1,obj2,elist);
}

ptr connectp(ptr obj1, ptr obj2, ptr elist)
{
  Reg ptr l,tmp;
  for(l=elist; CONSP(l); ) {
    tmp = CAR(l);
    if (CONSP(tmp)){
      if(equal(CAR(tmp), obj1) == Truth){
         if(equal(CDR(tmp), obj2) == Truth)
            return tmp;
      }
      if(equal(CAR(tmp), obj2) == Truth){
          if(equal(CDR(tmp), obj1) == Truth) 
          return tmp;
      }
    }
    if ((l=CDR(l)) == elist) goto Error;
  }
  if (NULLP(l)) return(Ntruth);
Error:
  Err("connection function: improper list", E_notlist,elist);
  return UNDEFINED; /* never reached */
}

ptr s_remove_vertex(int narg,ptr env)
{
  Reg ptr obj,  elist;
  Reg ptr l,x,tmp,z;
   checknarg(narg-2);
   obj=*(stack-2);
   elist=*(stack-1);

  tmp=z=CONS(NIL,NIL);
  psh(z);
  for(l=elist; CONSP(l); ) {
    x = CAR(l);
    if (CONSP(x)){
      if((equal(CAR(x), obj) == Truth)||(equal(CDR(x), obj) == Truth))
          ;
      else
        tmp=CDR(tmp)=CONS(x,NIL);
    }
    if ((l=CDR(l)) == elist) goto Error;
  }
  pull();
   return CDR(z);
Error:
  pull();
  Err("connection function: improper list", E_notlist,elist);
  return UNDEFINED; /* never reached */
}
ptr s_remove_edge(int narg,ptr env)
{
  Reg ptr obj,  elist;
  Reg  ptr l,x,tmp,z,rev;
   checknarg(narg-2);
   obj=*(stack-2);
   elist=*(stack-1);

  rev=CONS(CDR(obj),CAR(obj));
  tmp=z=SCONS(rev,NIL);
  psh(z);
  for(l=elist; CONSP(l); ) {
    x = CAR(l);
    if (CONSP(x)){
      if((equal(x, obj) == Truth)||(equal(x,rev) == Truth))
          ;
      else
        tmp=CDR(tmp)=CONS(x,NIL);
    }
    if ((l=CDR(l)) == elist) goto Error;
  }
  pull();
   return CDR(z);
Error:
  pull();
  Err("connection function: improper list", E_notlist,elist);
  return UNDEFINED; /* never reached */
}

ptr s_move_vertex(int narg, ptr env)
{
  Reg ptr obj1, obj2,  elist;
  /** obj1 -> obj2 **/
  Reg ptr l,x,tmp,z;
   checknarg(narg-3);
   obj1=*(stack-3);
   obj2=*(stack-2);
   elist=*(stack-1);

  tmp=z=CONS(NIL,NIL);
  psh(z);
  for(l=elist; CONSP(l); ) {
    x = CAR(l);
    if (CONSP(x)){
      if(equal(CAR(x), obj1) == Truth)
         tmp=CDR(tmp)=CONS(CONS(obj2,CDR(x)),NIL);
      else if(equal(CDR(x), obj1) == Truth)
         tmp=CDR(tmp)=CONS(CONS(CAR(x),obj2),NIL);
      else
        tmp=CDR(tmp)=CONS(x,NIL);
    }
    if ((l=CDR(l)) == elist) goto Error;
  }
  pull();
   return CDR(z);
Error:
  pull();
  Err("connection function: improper list", E_notlist,elist);
  return UNDEFINED; /* never reached */
}

ptr s_edge_assoc(int narg,ptr env)
{
  Reg ptr obj,  alist;
 /** obj(no-priented)  assoc alist **/
  Reg  ptr l,x,z,a,b;
    checknarg(narg-2);
    obj=*(stack-2);
    alist=*(stack-1);
  a=CAR(obj);
  b=CDR(obj);
  for(l=alist; CONSP(l); ) {
    x = CAR(l);
    if (CONSP(x)){
      z=CAR(x);
      if((equal(CAR(z), a) == Truth)&&(equal(CDR(z), b) == Truth))
         return x;
      if((equal(CAR(z), b) == Truth)&&(equal(CDR(z), a) == Truth))
         return x;
    }
    if ((l=CDR(l)) == alist) goto Error;
  }
   return Ntruth;
Error:
  Err("connection function: improper list", E_notlist,alist);
  return UNDEFINED; /* never reached */
}

ptr s_reduce_graph(int narg,ptr env)
{
  Reg ptr elist;
  /** (1 . 1)  -> remove,  double -> single  **/
  Reg  ptr x,l,tmp,z;
    checknarg(narg-1);
   elist=*(stack-1);

  z=tmp=CONS(NIL,NIL);
  psh(tmp);
  
  for(l=elist; CONSP(l); ) {
    x = CAR(l);
    if (CONSP(x)){
      if(equal(CAR(x), CDR(x)) == Ntruth)
        {
          if(connectp(CAR(x),CDR(x), CDR(z)) == Ntruth)
             {
               tmp=CDR(tmp)=CONS(x,NIL);
             }
        }
    }
    if ((l=CDR(l)) == elist) goto Error;
  }
  pull();
   return CDR(z);
Error:
  pull();
  Err("connection function: improper list", E_notlist,elist);
  return UNDEFINED; /* never reached */
}



