/***   03.12.29
    heap.c
    Mymalloc, Myfree,Myrealloc
   07.9.24  revised
 2019.10.14 New 
   type=FREE,USED,END,START
     Align is  head_size  unit
     heap is
       [START][BLOCK1]...[END]
     each BLOCK begins a heap_head
         [type size prev next]
     size is head_size of the Block (include head itself)
     START block has the size of TOTAL HEAP
     END block has  size=1
     heap[k] is a FREE list (bi-direction). by heap_size[k].
     Not call gbc
     see (malloc-info)
   2019.11.16  V2.18  heap_head==cell
***/
#define HEAP_C

#include  "lisp.h"

#include <stdio.h>
#include <string.h>
//#include <math.h>
#include <stdlib.h>

//#define DEBUG_HEAP
#ifdef  DEBUG_HEAP
#define INCTAB  debug_tab++;
#define DECTAB  debug_tab--;
#else
#define INCTAB  
#define DECTAB  
#endif
//  V2.18  heap_head is same as cell
/***
typedef struct heap_head {
   int type;
   int size;   // unit of HEAD, iclude the header itself
   struct heap_head *prev; // prev of heap list , bidirectional 
   struct heap_head *next; // make a heap list  
} HEAPHEAD,*headp;
***/
   /***  next is a block except for END-block  ***/

typedef  unsigned int  hsize;  /***  headp size  ***/
typedef  unsigned int  bsize;  /***  byte size   **/
typedef  ptr headp;  /***     **/
typedef  cell HEAPHEAD;  /***     **/

/*** type **/
#define FREE 0
#define USED 1
#define CELLU 2
#define END   4
#define START  8
#define CHECK   0x1000   // for free block check
/*** #define FREEMAX 32000   256K  ***/
#define HEAPMAX 21    //29
#define HTYPE(p) (((headp)p)->_attr)
#define HSIZE(p) (((headp)p)->_exattr)
#define HPREV(p) CAR(p)
#define NEXT(p)  CDR(p)
#define BASENEXT(p,s) ( (headp)(p)+s)
#define ALLOC_ADDR(p) ( (headp)(p)+1)
#define GET_HEADP(p) ( (headp)(p)-1)
#define head_size    (sizeof(cell))

static headp heap[HEAPMAX];
static headp startblock=NULL;  /** start block list this list is a malloc list!  **/
/**  use malloc for free table **/
//static int AL_SIZE=100000;  // 2.4M for SYSTEM64 
//static int AL_SIZE=70000;  /** 1.6M for SYSTEM64 ***/
static int AL_SIZE=80000;  /** 1.9M for SYSTEM64 ***/
//static int AL_SIZE=60000;  /** 1.4M for SYSTEM64 ***/
/***  static int AL_SIZE=30000;  **  for test ***/
static int heap_num=0;
static int debug_heap=0;
static int debug_tab=0;
int heap_check_sub(headp,int );
headp get_free(hsize size);
int  into_heap(headp p,int n);
//static int gbc_malloc=0;

#define isfree(p)  (HTYPE(p)==FREE)

headp new_heap(hsize,int);
void delete_free();
void delete_record();
void free_insert();
void ss_malloc_error();  /*** conflict to System of MACOSX  ***/
headp div_block();
int heap_check();
int malloc_info(char *);
hsize free_add();

void out_tab(){
  int n=debug_tab;
   while(n--){
     fputc(' ',stderr);
     fputc(' ',stderr);
  }
}

static void init_table();

void init_heap(){
 int i;
 for(i=0;i<HEAPMAX;i++) heap[i]=NULL;
 init_table();
 new_heap(0,FREE);
 //gbc_malloc=0;
}

void Myfree(headp p){
  headp b;
  headp next;
  hsize size;
   if(p==NULL) return;
   b=GET_HEADP(p);   // b is the head of p
   if(HTYPE(b)!=USED){
      fprintf(stderr,"illegal head type=%d p=%p \n",HTYPE(b),p);
      if(HTYPE(b)==FREE) return;
       ss_malloc_error(E_free);
   }
   size=HSIZE(b);
   next=BASENEXT(b,size);
#ifdef DEBUG_HEAP
     if(debug_heap) {
           out_tab();
           fprintf(stderr,"Myfree %p base block=%p hsize=%d\n",p,b,size);
      }
#endif
   if(isfree(next)){
     delete_free(next);
     size=free_add(b,next);
#ifdef DEBUG_HEAP
     if(debug_heap){
           out_tab();
         fprintf(stderr,"Myfree delete and add block %p total %d \n",next,size);
       }
#endif
     // HSIZE(b)=size;  by  free_add
   }
  into_heap(b,size);
}
/*** GMP requires old  **/
void MyfreeGMP(headp p,int old){
    Myfree(p);
}

hsize free_add(headp b,headp free){
 int s,k;
   s=HSIZE(b);
   k=HSIZE(free);
   s+=k;
   HSIZE(b)=s;
   return s;
}

hsize halign(bsize n){
  return 2+(n-1)/head_size; //OK
}

void *Mymalloc(bsize n){
  hsize size,get_size;
 headp new;
  /*** printf("Mymalloc %d\n",n);  ***/
  size=halign(n);
     //debug_tab++;
     INCTAB
  new=get_free(size);
  if(new==NULL){
      /**  printf("gbc from Mymalloc\n");  ***/
      /*** Never call gbc from here , it may cause gbc infinitely!!! ***/
      /** mada! gbc  ***/
     //gbc_malloc=1; 
     //fprintf(stderr, "CALL new_heap size=%d from Mymalloc\n",size); 
     new_heap(size,FREE);
     new=get_free(size);
     if(new==NULL) ss_malloc_error(E_malloc);
  }
  get_size=HSIZE(new);
#ifdef DEBUG_HEAP
  if(debug_heap){  
           out_tab();
      fprintf(stderr,"Mymalloc get_size %d  align=%d %p\n",get_size,size,new); 
     }
#endif
    HTYPE(new)=USED;
 if(get_size>=size+3){   // 4,5 makes many error  3 is org
   div_block(new,size);
 }
     //debug_tab--;
     DECTAB
   return ALLOC_ADDR(new);
}

unsigned int CELLU_size;
void *Mymalloc_CELL(bsize n){
  hsize size,get_size;
 headp new;
  // fprintf(stderr,"CELL Mymalloc bytes  %d\n",n);  
  size=halign(n);
     new=new_heap(size,CELLU);
  get_size=HSIZE(new)-1;
  // fprintf(stderr,"CELL Mymalloc  get_size %d  align=%d\n",get_size,size); /***/
   HTYPE(new)=CELLU;
   CELLU_size=(get_size);  // *head_size/sizeof(cell);  
                           // now same but -1 for head
   return ALLOC_ADDR(new);
}

/***  p is Mymalloc adress return real usable bytes **/
int get_alloc_bytes(headp p){
    hsize size;
  headp base;
  base=GET_HEADP(p);
  size=HSIZE(base);
   return (size-1)*head_size;  //  -1 for head
}

// memcpy has a problem?
void HDcpy(headp p,headp q,hsize n){
  Reg long *r=(long *)p,*s=(long *)q,m;
  m=n*sizeof(HEAPHEAD)/sizeof(long);
   while(m>0){
        m--;
       *r++=*s++;
   }
}
  

/***   realloc  ***/
void *Myrealloc(headp p,bsize new){
    hsize size,n;
  headp base,q,r;
  if(p==NULL) {
      return Mymalloc(new);
   }
  base=GET_HEADP(p);
  if(HTYPE(base)!=USED){
      fprintf(stderr,"illegal Realloc %d %p\n",HTYPE(base),p);
      ss_malloc_error(E_realloc);
  }
   n=halign(new);   /*** head_size align ***/
  size=HSIZE(base);
  if(n<=size){
     if(size<=2*n||size<=10) return p;
     HSIZE(base)=n;
#ifdef DEBUG_HEAP
     if(debug_heap) fprintf(stderr,"Myrealloc to  small  %p  hsize=%d from hsize %d\n",base,n,size);
#endif
     q=(headp)(base+n);
     HTYPE(q)=FREE;
     n=size-n;
     HSIZE(q)=n;
      //printf("Myrealloc free  %p  hsize=%d\n",q,n);
     //HPREV(q)=base;
     into_heap(q,n);  //Myfree(ALLOC_ADDR(q));
     return p;
   }
   n=(size-1)*head_size;
  r=Mymalloc(new);
#ifdef DEBUG_HEAP
   if(debug_heap){
        fprintf(stderr,"Myrealloc free  block %p  hsize=%d",base,size);
        base=GET_HEADP(r);
        fprintf(stderr," and new block %p  hsize=%d  realloc-adress %p\n",base,HSIZE(base),r);
   }
#endif
  memcpy(r,p,n);
  //  HDcpy(r,p,size-1);
  Myfree(p);
  return (void *)r;
}

/***   realloc , the term  old is for GMP ***/
void *MyreallocGMP(headp p,bsize old,bsize new){
  hsize size,n,h_old;
  headp base,q,r;
  //char *r;
  if(p==NULL) {
      return Mymalloc(new);
   }
  base=GET_HEADP(p);
  if(HTYPE(base)!=USED){
      fprintf(stderr,"illegal Realloc %d %p\n",HTYPE(base),p);
      ss_malloc_error(E_reallocgmp);
  }
   n=halign(new);   /*** head_size align ***/
  h_old=halign(old);   /*** head_size align ***/
  size=HSIZE(base);
  if(size<h_old){
      fprintf(stderr,"illegal Realloc old %d %p\n",old,p);
      ss_malloc_error(E_reallocgmp);
   }
  if(n<=size){
     if(size<=2*n||size<=10) return p;
     HSIZE(base)=n;
      //printf("Myrealloc to  small  %p  hsize=%d from hsize %d\n",base,n,size);
     q=(headp)(base+n);
     HTYPE(q)=FREE;
     n=size-n;
     HSIZE(q)=n;
      //printf("Myrealloc free  %p  hsize=%d\n",q,n);
     //HPREV(q)=base;
     into_heap(q,n);  //Myfree(ALLOC_ADDR(q));
     return p;
   }
  if(old==0) old=size*head_size;
  r=Mymalloc(new); 
  //if((headp)r==p) return p;  
  //memcpy(r,p,old);
    HDcpy(r,p,h_old-1);
   /***
  sr=(char *)p;
  base=(headp)r;
  while(old--) *r++=*sr++;
    ***/
  Myfree(p);     
  return (void *)r;
}

headp div_block(headp free,hsize n){
 hsize s;
 headp new;
    s=HSIZE(free);
    HSIZE(free)=n;
   new=BASENEXT(free,n);
   HTYPE(new)=FREE;
   s-=n;
   HSIZE(new)=s;
   into_heap(new,s);   
   return new;
}

int size_heap[HEAPMAX];
/***
int Log2(int n){
  ***  n=2  0, =3 1, =4  2,<=8 3,<=16 4,... ***
  ***  change  **
  int k,g;
   if(n==2) return 0;
   if(n==3) return 1;
   k=2;
   g=4;
  while(g<n){
    //g*=1.5;
    g*=1.25;
    k++;
 }
 if(k<HEAPMAX) return k;
 return HEAPMAX-1;
}
***/
int Log2(int n){
  int k,g;
   if(n<=8) return n-2;
  for(k=6;k<HEAPMAX-1;k++){
   g=size_heap[k+1]; // n=9 return 6
   if(n<g) return k;
  }
  return k;
}
static void init_table(){
  int k,g;
  size_heap[0]=2;
  size_heap[1]=3;
   k=2;
   g=4;
   //printf("heap[%d]=%d\n",0,2);
   //printf("heap[%d]=%d\n",1,3);
  while(k<HEAPMAX){
     size_heap[k]=g;
  // printf("heap[%d]=%d\n",k,g);
    if(k>10) g*=2;
    else g*=1.25;
    k++;
  }
}
   


int  into_heap(headp p,int n){
  /*** not sort ***/
  headp q;
  int k;
  HTYPE(p)=FREE;
  k=Log2(n);
  q=heap[k];
#ifdef DEBUG_HEAP
   if(debug_heap){ 
     INCTAB
           out_tab();
     DECTAB
     fprintf(stderr,"into heap block %p k=%d hsize=%d  org heap[%d] %p\n", p,k,n,k,q);
     }
#endif
  NEXT(p)=q;
  HPREV(p)=NULL;
  if(q!=NULL){
    HPREV(q)=p;
  }
  heap[k]=p;
 return k;
}

headp get_free(hsize size){
   headp p,q,r;
   int k,n;
   k=Log2(size);
   //printf("get_free size %d  k=%d\n",size,k);
   while(k<HEAPMAX){
    p=heap[k];
    while(p!=NULL){
       if(HSIZE(p)>=size){
           q=HPREV(p);
           if(q!=NULL){
             r=NEXT(p);
             NEXT(q)=r;
             if(r!=NULL)  HPREV(r)=q;
           }
           else{
             q=NEXT(p);
             heap[k]=q;
             if(q!=NULL)  HPREV(q)=NULL;
           }
#ifdef DEBUG_HEAP
  if(debug_heap) { 
           out_tab();
    fprintf(stderr,"get_free return %p for %d heap[%d]\n",p,size,k);
           out_tab();
    fprintf(stderr,"get_free Now  heap[%d]=%p\n",k,heap[k]);
     }
#endif
           return p;
       }
       p=NEXT(p);
    }
   k++;
  }
#ifdef DEBUG_HEAP
  if(debug_heap) { 
           out_tab();
    fprintf(stderr,"get_free return NULL\n");
  }
#endif
  return NULL;
}

  
headp new_heap(hsize n,int type){
  headp p;
  headp end;
  int by;
/** printf("new heap start n  %d\n",n);  **/
  heap_num++;
  if(heap_num>4) AL_SIZE+=100000;
  if(AL_SIZE>n) n=AL_SIZE;
  //printf("new heap malloc %d\n",n*head_size);
  n+=3;   // 3 head can not use
  by=(n)*head_size;
  if(by<n) {
     printf("new_heap overflow request hsize %d by=%d\n",n,by);  
     ss_malloc_error(E_malloc);// overflow request 
   }
  p=malloc(by);
  if(p==NULL) {
     printf("new_heap malloc error request hsize %d by=%d\n",n,by);  
     ss_malloc_error(E_malloc);// overflow request 
   }
  HTYPE(p)=START;
#ifdef DEBUG_HEAP
if(debug_heap){
           out_tab();
    fprintf(stderr, "new start %p %d  n %x end %p \n",p,START,n,(p)+n);
 }
#endif
  NEXT(p)=startblock;
  startblock=p;
  HSIZE(p)=n;   /*** Caution!  start block has total size  ***/
  end=BASENEXT(p,n-1);
  HTYPE(end)=END;
  HSIZE(end)=1;
  HPREV(end)=p;
  p++;
  HTYPE(p)=type;
  HSIZE(p)=n-2;
     INCTAB
  if(type==FREE){
         into_heap(p,n-2);
  }
#ifdef DEBUG_HEAP
     DECTAB
if(debug_heap)
      malloc_info("new heap");
#endif
  return p;
 }

void delete_free(headp b){
 headp s,q;
 hsize n;
 int k;
   n=HSIZE(b);
#ifdef DEBUG_HEAP
     if(debug_heap) {
           out_tab();
        fprintf(stderr,"delete_free %p hsize %d\n",b,n);
     }
#endif
   s=HPREV(b);
   q=NEXT(b);
 //printf("delete_free  prev %p next %p \n",s,q);
   if(s==NULL){
     k=Log2(n);
#ifdef DEBUG_HEAP
     if(debug_heap){
           out_tab();
       fprintf(stderr,"delete_free change heap[%d] to %p hsize %d\n",k,q,n);
     } 
#endif
    heap[k]=q;
   }
   else{
     NEXT(s)=q;
   }
  if(q!=NULL) HPREV(q)=s;
}


void ss_malloc_error(int n){
 fprintf(stderr,"Malloc error %d\n",n);
 s_bye(0,a_nil);
 //fatal_error(E_malloc,a_nil);
}  

static int errheap=0;
       
int heap_check(){
  int i,s=0;
  for(i=0;i<HEAPMAX;i++){
    s+=heap_check_sub(heap[i],i);
  }
       if(errheap>0){
            printf("error heap errct=%d\n",errheap);
            ss_malloc_error(E_heap);
       }
       printf("total free blocks %d\n",s);
 return s;
}

int heap_check_sub(headp p,int  k){
 int n,s,ct=0,pos=0;
 headp prev=NULL;
 while(p!=NULL){
   n=HTYPE(p);
   s=HSIZE(p);
   if(HPREV(p)!=prev){
       printf("illegal Prev  heap[%d] type %d %p pos=%d\n",k,n,p,pos);
       errheap++;
   }
   if(n!=FREE){
       printf("illegal HEAD TYPE heap[%d] %d %p pos=%d\n",k,n,p,pos);
       errheap++;
     }
   else if(s==0){
       printf("illegal hsize=0  k= %d p=%p pos=%d\n",k,p,pos);
       errheap++;
  } else
     ct++;
     pos++;
     prev=p;
     p=NEXT(p);
   }
 return ct;
}

int heap_mark(){
  int i,s=0;
  headp p;
  for(i=0;i<HEAPMAX;i++){
    p=heap[i];
     while(p!=NULL){
      HTYPE(p)=CHECK;
       s+=HSIZE(p);
       p=NEXT(p);
     }
  }
 return s;  // return total hsize
}

static int free_total;
static int used_ct[HEAPMAX];

int free_check(int msg){  //all and only free, no message if normal
  headp p=startblock,np;
  int n,s,a,err=0,total;
   total=heap_mark();
   free_total=total;
  for(n=0;n<HEAPMAX;n++) used_ct[n]=0;
  while(p!=NULL){
   n=HSIZE(p);
   s=HTYPE(p);
   if(s!=START){
      if(msg) printf("Abnormal system first %p type=%d hsize=%d\n",p,s,n);
      err++;
      break;
     }
   else{
      if(msg) printf("malloc start block  %p type=%d hsize=%d\n",p,s,n);
    }
   np=NEXT(p);// next startblock
   p=p+1;
   while(p!=NULL){
     n=HSIZE(p);
     s=HTYPE(p);
     if(n==0){
              err++;
              if(msg) printf("Abnormal size 0  block %p   type %d\n",p,s);
              break;
     }
      if(s==END) break;
      else if(s==CHECK){  //OK this is FREE  by mark
         HTYPE(p)=FREE;
         total-=n;
      }else if(s==FREE)
          {
              err++;
              if(msg){ 
                 printf("Free but not in heap %p  hsize %d\n",p,n);
              }
          }
      else if(s==CELLU){
              if(msg) printf("  Cell bank %p  hsize %d\n",p,n);
      }
      else if(s==USED){
        s=Log2(n);
        used_ct[s]++;
      }
      else{
              err++;
              if(msg) printf("Abnormal type block  %p  hsize %d  type %d\n",p,n,s);
              break;
      }
     p+=n;
   }
   p=np;
 }
  if(total){
           err++;
           if(msg)   printf("Free total %d is different %d\n",free_total,total);
  }
 return err;
}
  
int alloc_check(){  //all
  int k,c,err,total;
  heap_check();
    err=free_check(1);
   total=free_total;
  if(err==0) printf("Free total hsize %d  correct!\n",total);
   printf("Used block count\n");
   for(k=0;k<HEAPMAX;k++){
        c=used_ct[k];
    if(c) printf("k=%d: %d   hsize %d\n",k,c,size_heap[k]);
   }
   printf("alloc_check end\n");
 return err;
}

/****   for temporaly get strage   size  by  ptr   04.6.12  for set.c  ***/
ptr *ptr_alloc(int n,ptr init)
{
  Reg ptr *p;
  p=(ptr *)Mymalloc(n*sizeof(ptr));
  if(init){
    p+=n;
    while(n--) *--p=init;
  }
  return p;
}
/***  from s_bye  ***/
void free_sys(){
  headp p=startblock,q;
  while(p!=NULL){
     q=NEXT(p);
     free(p);
     p=q;
  }
//  printf("free_sys end\n");
}
#ifndef MAIN
ptr s_malloc_info(int narg,ptr env){
 int s;
 checknarg(narg);
 s=malloc_info("");
 return MAKESINT(s);
}
ptr s_malloc_debug(int narg,ptr env){
   if(narg>0&&*(stack-1)==a_f)
     debug_heap=0;
    else debug_heap=1;
     return PRED(debug_heap);
}

#endif
static int hlength(headp p){
  int c=0;
  while(p!=NULL){
     c++;
     p=NEXT(p);
   /***
     if(c>10000){
  printf("heap length over %d   exit\n",c);
       exit(999);
       break;
     }
   ***/
  }
  return c;
}


int malloc_info(char *mes){
 headp p;
 int i,ct,s;
  printf("malloc_info start %s\n",mes);
  for(i=0;i<HEAPMAX;i++){
    p=heap[i];
   if(p!=NULL){
    ct=hlength(p);
    s=HSIZE(p);
    printf("heap:%d ct=%d  p %p  hsize %d\n",i,ct,p,s);
   }
 }
  // heap_check(); ->alloc_check
  i=alloc_check();
  if(i) {
       ss_malloc_error(E_heap);
    }
  //printf("\n hsize %lu  cell size %lu\n",head_size,sizeof(cell));
  printf(" malloc_info end\n");
  return 0;  /** toriaezu **/
}

