/***   timer.c 
   timer_handler 
    04.07.10
    04.7.22  revised
    see note
    05.4.14   (after 100)
    14.5.23 
    14.11.24  absolute time (by msec)
***/

#include "lisp.h"
#include <sys/time.h>
#include <signal.h>
#include <stdio.h>


/** #define DEBUG 1 **/
static struct itimerval value,ovalue;
static int timer_int,max_after,after_total;
static int after_entry[AFTERMAX]; 
       /***  positive  interval( active) ,-1 not entry, 0 go state ***/
static int Lock_after;
static int Lock_exec;
/*** 14.11.24 absolute time by msec ***/
static int sstime=0;
static int first_event;


#define  SECMILI  2000    /****  Linux 1 sec is 2sec=2000 milisec in real time  ***/ 
#define  MICROMILI  500    
/****  Linux 1 micro is 2microsec=2/1000=1/500  milisec in real time***/ 

ptr after_cmd[AFTERMAX];   /***  needs gbc  ***/

void exec_check();
void timer_check();
ptr cmd_exec();
ptr s_update();

/*** init  ***/
void timer_init()
{
  Reg int i;
     /*** after_cmd=NIL;   **  needs gbc  ***/
  for(i=0;i<AFTERMAX;i++){
    after_entry[i]=-1;
    after_cmd[i]=a_nil;
  }

  after_total=0;
  timer_int=0;  /** not active ***/ 
  first_event=-1;
   Lock_after=0;
   Lock_exec=0;
}

int SetTimer(int msec){
    int sec,usec;
    timer_int=msec;
    sec=msec/1000;
    usec=1000*(msec-sec*1000);
    //  timer Not restart implicitly
    value.it_interval.tv_sec= 0;  //sec 
    value.it_interval.tv_usec= 0;  //usec
    //
    value.it_value.tv_sec= sec;
    value.it_value.tv_usec= usec;
    if (setitimer(ITIMER_REAL, &value, NULL) != 0)
      fprintf(stderr,"setitimer Fail!\n");
#ifdef DEBUG
    fprintf(stderr,"SetTimer msec=%d\n",msec);
#endif
    return 0;
}
//  return rest time by msec
int GetTimer(){
   int s;
   getitimer(ITIMER_REAL,&ovalue);
   s=ovalue.it_value.tv_usec/1000;
   return s+ovalue.it_value.tv_sec*1000;
}

/***  timer handler ***/
void timer_handler(int sig){
 //sys_alarm++;
  timer_check();
}

/*** timer_check should be re-entrant!!   05.4.22 ***/
void timer_check(){  /***  [1]3-5  ***/
 Reg int i,n,j,interval,min=1000000000;
 ptr cmd;
  interval=timer_int;
  if(interval>0)
        SetTimer(0);   //Stop!
#ifdef DEBUG
 fprintf(stderr,"enter timer check  interval %d alarm %d total %d\n",
    interval,sys_alarm,after_total);  
#endif
  after_total=0;

  sstime+=interval;
 //printf("sstime %d timer_int %d\n",sstime,interval);
  if(first_event>=0){
  cmd=after_cmd[first_event];
  after_entry[first_event]=-1;  //empty
  }
  else
    cmd=a_nil;
  first_event=-1;
 for(i=0;i<=max_after;i++){
    j=after_entry[i];
    if(j>=0){
       after_total++;
        if(j<=sstime){
#ifdef DEBUG
              printf("time come! %d\n",i); 
#endif
             after_entry[i]=0; /***  go state ***/
             after_total--;
             /*** cmd_exec(after_cmd[i]);  don't change after_cmd after this! ***/
                              /***  because cmd may call after!!! ***/
       }
      else if(j<min){
           min=j;
           first_event=i;
      }
      else
          ;
    }
 }
  if(first_event>=0){
    min-=sstime;
    if(min<=0) min=10;  // at least 10 msec
    SetTimer(min);   // start timer
  }
  /*** timer_check_start();  ***/
  Lock_exec=1;
  exec_check(cmd);  // cmd and others will exec
  Lock_exec=0;
#ifdef DEBUG
  printf("out timer check  after_total %d  min-interval %d \n"
       ,after_total,min); 
#endif
}

void exec_check(ptr cmd){
  int i,j;
  if(cmd!=a_nil)
     cmd_exec(cmd);
  
  for(i=0;i<=max_after;i++){
    j=after_entry[i];
    if(j==0){
#ifdef DEBUG
  printf("exec_check timer check  entry %d  d total %d\n",i,after_total);
#endif
       after_entry[i]=-1; /***  clear ***/
       cmd_exec(after_cmd[i]); /***  don't change after_cmd after this! ***/
       /*** after_cmd[i]=a_nil;  Don't do this !!!***/
     }
  }
}

/***   interval is past time && return of MIN of rest time  ***/
/***  set sstime=0 and change after_entry  ***/
int  timer_diff(){  /***  [2] 4-6  ***/
 Reg int i,n,min=sstime;
   SetTimer(0);
 for(i=0;i<=max_after;i++){
    if(after_entry[i]>0){
        n=after_entry[i]-sstime;
        if(n<0)  n=0;
        after_entry[i]=n;
    }
 }
 sstime=0;
 return 0;
}

static ptr make_after_entry(int n,ptr cmd) /***  **/
{
  Reg int i,rest=0,interval,newtime;
  for(i=0;i<AFTERMAX;i++){
    if(after_entry[i]==-1) break;
  }
  if(i==AFTERMAX){
      lwritestr(STDERR,"after cmds over MAX!\n");
      error(E_after,a_nil);
   }
  after_total++;
  interval=timer_int;
  if(timer_int>0)
      rest=GetTimer();
#ifdef DEBUG
  fprintf(stderr,"make_after rest=%d timer_int %d\n",rest,interval);
#endif
  SetTimer(0);  /** for test  **/
  newtime=sstime+interval-rest+n;
  after_entry[i]=newtime;
  /***  may be first_event  ****/
  if(first_event>=0){
    if(after_entry[first_event]>newtime){
    /** cont ***/   
      //rest=GetTimer();
      sstime+=interval-rest;
      //sstime+=past;
      first_event=i;
      SetTimer(n);
      //timer_int=n;  by SetTimer
    }
   else
      SetTimer(rest);
  }
  else{
    first_event=i;
    SetTimer(n);
  }
  after_cmd[i]=cmd;
  if(i>max_after) max_after=i;
  return MAKESINT(i);
}

static ptr after_wait(ptr cmd){
  /***  2014.12.1  wait for CDR(cmd)==a_t ***/
  while(CDR(cmd)==a_nil){
    s_update(0,a_nil);
  }
 return a_undef;
}

ptr s_after(int narg,ptr env)
{
 ptr sec,cmd,ent;
 int n,y,afsleep;
  if(narg==1){
  afsleep=1;
  sec=*(stack-1); 
  cmd=CONS(a_f,a_nil);
  }
  else  if(narg==2){
  sec=*(stack-2); 
  cmd=*(stack-1);
  afsleep=0;
  }
  else  checknarg(narg-2);

  if(!SINTP(sec)){
      lwritestr(STDERR,"interval < 500000000 (5 days)\n");
      error(E_range,sec);
  }
  n=GETNUM(sec);
  //printf("after %d\n",n);
  if(n<20) n=20;
   ent=make_after_entry(n,cmd); /*** entry  **/
   if(afsleep) return after_wait(cmd);
   return ent;    
}

void ss_update(){
 int i=10000;
if(Lock_exec) return;  /***  2014.11.29 **/
   if(timer_int==0) timer_check();
 if(sys_alarm==0){ 
   /*** for cygwin   SIGALRM taken by TCL in cygwin ****/
   //if(after_total&&(value.it_value.tv_sec==0)&&(value.it_value.tv_usec==0)){
   if(after_total&&(timer_int==0)){
       signal(SIGALRM,timer_handler);   /*** because TCL/TK take timer or block? ***/
       //while(i) i--;
    }
   else
       return;
  }
  /*** fprintf(stderr,"ss_update   alarm %d\n",sys_alarm); ***/ 
  /*** timer_stop();  ***/  
  /***  2014.11.29!!  **/
  /***  timer_check();   ***/
  /** printf("ss_update\n"); ***/ 
}
    
ptr cmd_exec(ptr p)
{
  if(SREGP(p)){
    return p;
  }
  if(CLOSUREP(p)){
     p=funcall(p,0,a_nil);
     return p;
  }
  if(CONSP(p)&&(CAR(p)==a_f)){
    CDR(p)=a_t;
    return p;
  }
  psh(p);
  /*** p=meval(p,a_nil);    1.99  ***/
  p=vm_eval(p,a_nil);  /***  V2.0   2010.5.30  ***/
  pull();
  return p;
} 

static ptr after_cancel(int i)
{
  int n;
  ptr cmd;
  n=after_entry[i];
  if(n==-1) return a_f;
  after_entry[i]=-1;
  cmd=after_cmd[i];
  after_cmd[i]=a_nil;
  after_total--;
  if(CONSP(cmd)&&(CAR(cmd)==a_f)) CDR(cmd)=a_t;
  if(i==first_event){
     first_event=-1;
       timer_check();
  }
  else if(after_total==0){
           SetTimer(0); /*** Never ocuur? ***/
  }
 else
    ;
 return a_t;
}

void after_info(){
  Reg int i,n,past,sec,time_rest;
   getitimer(ITIMER_REAL,&ovalue);
  // time_rest=ovalue.it_value.tv_usec/MICROMILI;
   time_rest=ovalue.it_value.tv_usec/1000;
   sec=ovalue.it_value.tv_sec;
   //time_rest+=sec*SECMILI;
   time_rest+=sec*1000;
   past=timer_int-time_rest;
    sprintf(iobuff,"interval %d time_rest %d past %d active %d sstime%d\n\n",
     timer_int,time_rest,past,after_total,sstime);  //timer_active

    lwritestr(STDOUT,iobuff);
    /** printf("sys_alarm %d\n",sys_alarm); ***/
  for(i=0;i<=max_after;i++){
     if(after_entry[i]>0){
#ifdef DEBUG
         prin1(MAKESINT(i),STDOUT);
         lwritec(STDOUT,':');
         //prin1(MAKESINT(after_timer[i]-past),STDOUT);
         prin1(MAKESINT(after_entry[i]-sstime),STDOUT);
         lwritec(STDOUT,' ');
         prin1(MAKESINT(after_entry[i]),STDOUT);
         lwritec(STDOUT,' ');
         prin1(after_cmd[i],STDOUT);
         newline(STDOUT);
#endif
       }
    }
 }

ptr s_after_cancel(int narg,ptr env)
{
 ptr sec;
 int n;

  checknarg(narg-1);
  sec=*(stack-1);
  n=GETNUM(sec);  
  if(n<0||n>=AFTERMAX) error(E_range,sec);
  return after_cancel(n);
}

ptr s_after_info(int narg,ptr env)
{
  checknarg(narg);
  after_info();
  return a_undef;
}

ptr s_update(int narg,ptr env)
{
  if(narg>1)
   checknarg(narg-1);
  ss_update();
  if(narg==1) return a_undef;
#ifdef USE_TK
  update();
  /*** signal(SIGALRM,timer_handler);   because TCL/TK take timer ***/
#endif
  return a_undef;
}


