;;; usage:  (get-file-list )
;;;         (get-file-list  "*.sgf")
;;;         (get-file-list  "*.scm" "*.stk" "*.lsp")
;;;  
;;;  if cancel,  ()  returned
;;;  else  a list of files  is returned
;;;
;;;      2002.12.24  
;;;      2005.3.15 revised
;;;      2005.3.21  Double click version
(require "get-file-base")
(in-package "get-file")
(define *get-file-list* ())
#|
(define fixfont "fixfont")
(let* ((name (font 'names)))
   (if (string? name) (set! name (list name)))
   (if  (not (member fixfont name))
    (font 'create fixfont :family 'fixed :size 11)
   )
)
|#

;;;;(defstruct GETF :cwd  :dirs :files :allfiles :win :dirname :var :globent)

(define *dir-list* ())
(define *dir-listing* ())
(define *file-ext* "")

(define (get-file-list  . ext)
  (let*  ((n-files 0) 
           (dir (gen-toplevel ".dir"))
           (wdir  (tk-name  dir ".wdir")) ;;;  file frame
           (labfiles (tk-name  dir ".labfiles"))
           (files (tk-name  dir ".files"))
           (files_ysc (tk-name  dir ".fiels_ysc"))
           (files_xsc (tk-name  dir ".fiels_xsc"))
          ;;; (ent (tk-name  dir ".ent"))
           (ok (tk-name  dir ".ok"))
           (add (tk-name  dir ".add"))
           (cancel (tk-name  dir ".cancel"))
           (remove (tk-name  dir ".remove"))
           (clear (tk-name  dir ".clear"))
           (getf (make-GETF ))
          )
     (set!  *get-file-list* ())
    (if (null? ext)   (set! ext '("*") ) )
    (set! *file-ext* ext)
    (wm 'title  dir  "File Multi Select window")
    ;;;(wm 'geometry dir "790x430+80+80")
    (frame wdir)
     (set-GETF getf :top wdir :ext ext)
    ;;;(set! getf (make-files-win wdir ext))
    (make-files-win getf)
    (funcall (GETF-win getf) 'configure :selectMode 'extended)
     (label labfiles :text "Selected files")
     (grid labfiles :row 0 :column 1)

      (listbox  files  :selectMode 'extended
        :yscrollcommand
                  (lambda l (apply  (function files_ysc)  'set l))
         :xscrollcommand
                  (lambda l (apply  (function files_xsc) 'set l))
       :width 65 :height 16)
     (scrollbar  files_ysc  :orient 'vertical
       :command   (lambda l (apply  (function files)  'yview l))
     )
     (scrollbar  files_xsc :orient 'horizontal
       :command   (lambda l (apply  (function files)  'xview l))
     )

     (grid  wdir   :row 2  :sticky 'nws)
     (grid    files :row 2  :sticky 'news :column 1 :columnspan 2)
     (grid  files_ysc  :row 2  :sticky 'nws :column 3 )
     (grid  files_xsc :column 1 :columnspan 2 :sticky 'ew)
   ;;;  (entry ent )
     (button ok :text "OK" 
         :command (lambda() (get-list-files files)(destroy dir)))
     (button add :text " Add ->" :command (lambda()(add-files getf files )))
     (button remove :text "remove"  :command (lambda()(sel-remove files)))
     (button clear  :text "All clear"  :command (lambda() (all-clear files)))
     (button cancel :text "CANCEL" :command 
         (lambda() (set! *get-file-list* ())(destroy dir))  )
    ;;; (grid ent  :column 0 :sticky 'ew ) 
     (grid add :row 4 :column 0 :sticky 'e ) 
     (grid cancel  :row 5 :column 1 :sticky 'e ) 
     (grid remove  :row 4 :column 2 :sticky 'w ) 
     (grid clear  :row 5 :column 2 :sticky 'w ) 
     (grid  ok :row 4 :column 1 :sticky 'ew )   ;;;  :sticky 'ew )
     (funcall files 'xview 'moveto 1.0)
      (dir-reset getf)
    (select-reset files) 
    ;;;(bind  ent  "<Return>" (lambda () (ent-chdir file-lb wdir labwdir ent)))
     (grid 'rowconfigure dir 2 :weight 1)
     (grid 'columnconfigure dir 1 :weight 1)
    (tkwait 'window dir )
         *get-file-list*
))

;;;;;
(define (select-reset files)
     (do ((x ()) (i 0) (file-list *get-file-list*))
          ((null? file-list) i )
         (set! x (pop file-list))
          (funcall files 'insert 'end  x)
          ; (format #t "~a~%" x)
          (set! i (+ i 1))
       )
)

(define (add-files getf files)
 (let* ((lb (GETF-win getf))
         (allf (GETF-allfiles getf))
        )
  (do ((i 0) (file 0)(top (getcwd))
          (all(listbox-index lb ))(res (reverse *get-file-list*)))
      ((null? all)  (set! *get-file-list* (reverse res)))
      (set! i (pop all))
      (set! file (list-ref allf i))
      ;;;(set! file (tk-string top "/" (funcall lb 'get i)))
      (set! file (tk-string top "/" file))
      (set! res (cons file res))
      (funcall files 'insert 'end file)
  )
(funcall files 'xview 'moveto 1.0)
))

;;(get-file-name)
(define (list-clear files)
  (funcall files 'delete 0 'end)
)
(define (list-e-reset files)
     (do ((x ()) (i 0) (file-list  *get-file-list*))
          ((null? file-list) i )
         (set! x (pop file-list))
          (funcall files 'insert 'end  x)
          ; (format #t "~a~%" x)
          (set! i (+ i 1))
       )
   (funcall files 'xview 'moveto 1.0)
)

(define (get-list-files files)
  (set! *get-file-list* (funcall files 'get 0 'end))
)

(define (sel-remove files)
   (let* ( (file 0)(en 0)
          (all(listbox-index files ))(res() ))
       (set! en (funcall files 'index 'end))
    (when(pair? all)
       (sort all <)
       (do ((i 0)(x (pop all)))
         ((= i en) (set! *get-file-list* (reverse res)))
         (if (= i x)
            (if (pair? all) (set! x (pop all)))
           ;;;else
           (push (funcall files 'get i) res)
          )
         (set! i (+ 1 i))
       )
    )
 (list-clear files)
 (list-e-reset files)
))

(define (all-clear files)
   (set! *get-file-list* ())
   (list-clear files)
)

(export "get-file-list")
(use-package "get-file" "ss")


