;;;  regexp2.0
;;;  R-reduce.ss
;;;   2005.8.19
;;;

;;;  reduce rules
;;;   :e  == epsilon transition
;;;  

(in-package "regexp")

(define *state* ())
(define *state-num* 0)
(define *state-trans* ())
(define *fail* '(FAIL))


(define (new-state exp)
   (if (number? exp)
       exp
     (let* ((n (assoc exp *state*))(x *state-num*))
       (if n
         (cdr n)
         (begin
          (inc *state-num*)
          (push (cons exp x) *state*)
          x
        )
      )
    )
  )
)

(define (force-new-state exp)
     (let* ((x *state-num*))
          (push (cons exp x) *state*)
          (inc *state-num*)
     x
))


(define (get-link lnk)
   (let* ((E (rassoc lnk *state-trans*)))
     (if E
         (car E)
         lnk
     )
))

(define (set-trans E red)
   (let* ((trans (assq E *state-trans*)))
     (if (and trans (equal? red (cdr trans)))
         red
         (push (cons E red) *state-trans*)
    )
))

(define (R-reduce exp)
   (set!  *state-num* 0)
   (set!   *state* ())
   (set!  *state-trans* ())
   (new-state ())  ;;;;   epsilon is 0
   (R-red  exp )  ;;; ORG=1
)
  
(define (R-red exp )
      (cond ((null? exp) 0)
            ((number? exp) exp)
            (else (R-red-pair exp ))
      )
)

(define (R-red-pair exp )
   (let* ((x (car exp)) (rest (cdr exp)))
      (if (number? x)
         (R=red-num exp x rest)
         (R-red-pair-sub exp  x rest)
      )
))

(define (R=red-num exp x rest)
   (let* ((E (new-state exp)) (lnk 0)(B 0))
        ;;; (format #t "R=red-num exp ~s x ~s  rest ~s E ~s~%" exp x rest E)
          (if (eq? x 0)
               (begin
                  (set! lnk (R-red rest ))   ;;;; ( 0 .  B) -> B
                  (set-trans E (get-link lnk))
                  lnk
               )
             (if (number? rest)
                (if (eq? rest 0)
                     (begin
                        (set-trans E x)
                         x     ;;;;  (x . 0) -> x
                      )
                    (begin
                       (set-trans E exp)
                        exp
                     )
                  )
                 (begin
                   (set! lnk (R-red rest))
                   (set!  B (get-link lnk))
                   (set! lnk (cons x B))
                   (set-trans E lnk)
                   lnk
                 )
               )
           )
))

(define (concate a b)
   (if (pair? a)
        (append a b)
        (cons a b)
   )
)

(define (R-red-pair-sub exp x rest)
   (let* ((E (new-state exp))(top (car x))(lnk 0) (tr 0) (f 0)(F 0)
            (A 0)  (c 0)(d 0)(C 0)(D 0) (n 0) (m 0)
           )
     ;;; (format #t "red-pair-sub exp ~s ~% top ~s ~% x ~s ~% rest ~s~%" exp top x rest)
    (set! tr
      (cond ((symbol? top)
                      (set! A (new-state x))  ;;;  xexp  maybe equal exp  
                      (set-trans A x)     ;;;;  atomic pattern trans not exp!
                      (set! lnk (R-red rest))
                       (cons A (get-link lnk))
              )
            ((eq? top :or) 
                  (R-red-or top x rest )
               )
             (  (eq? top :*+)
                   (R-atomic-pat (cons :* (cdr x)) rest)
               )
             ( (or (eq? top :*) (eq? top :*?)) ;;;  (r-*  exp)
                    (R-red-* top x rest )
               )
             ((eq?  top :++)
                   (R-atomic-pat (cons :+ (cdr x)) rest)
              )
             ((or (eq?  top :+) (eq?  top :+?))
                      (R-red-+ top x rest )
               )
             ((eq? top :back)  ;;;  (:back  exp)  modoriyomi
                  (R-red-back top x rest E)
               )
              ;;; New!
              ((eq? top :backward)  ;;;  (:backward  exp)  modoriyomi
                  (R-red-backward  top x rest E)
               )

              ((or (eq? top :range) (eq? top :range?))
                   (R-red-range top x rest)
               )
             (  (eq? top :range+)
                   (R-atomic-pat (cons :range  (cdr x)) rest)
               )
            ((or (eq? top :?) (eq? top :??))
                 (R-red-?? top x rest)
               )
            ((eq? top :?+)
                   (R-atomic-pat (cons :?  (cdr x)) rest)
              )
           ((or (eq? top :not) (eq? top :forward))  ;;;;;(eq? top :try))   ;;; :not
                (R-red-not top x rest)      
           )    
            ((eq? top :atomic)   ;;; 
                (R-red-atomic top x rest)
           )
           ((eq? top :if)   ;;; :if
                (R-red-if top x rest)
           )
      )
    )
   (set-trans E tr)
    ;;;(format #t "set-trans E ~s tr ~s exp ~s~%" E tr exp)
   tr
))

(define (R-red-or top x rest )
   (let* ((c 0)(C 0)(d 0)(D 0))
     (set! c (cadr x))
                (set! d (caddr x))
                (when (not (null? rest))   ;;; rest may be a Number!
                    (set! c (concate c rest))
                    (set! d (concate d rest))
                )
                (set! C (R-red c))
                (set! D (R-red d))
                 (list  (list :or (get-link C)(get-link D)))
))


(define (R-red-back top x rest E)
   (let* ((exp 0)(c 0)(C 0)(d 0)(D 0))
;;;;;(format #t ":back  x ~s ~%" x)
          (set! c (cadr x))
          (set! d (pop c))
          (set! exp   (list (list :or (append  c rest) (cons  d E))))
            (R-red  exp)
))

(define (R-red-backward  top x rest E)
   (let* ((exp 0)(c 0)(C 0)(d 0)(D 0))
;;;;;(format #t ":back  x ~s ~%" x)
                 (set! c (cadr x))
                 (set! c (backward-trans c))
                 (set! C (R-red c))
                (set! D (R-red rest))
      (cons (get-link C)(get-link D))
))

(define (backward-trans c)
   (let* ((f (list 0)) (res (list (list 'r-pos f))) )
       (set! res (append c res))  
       (push (list 'inc-chk-pos f) res)
       (set! res (list (list :back res)))
       (push (list 'back-pos (R-count c) f) res)
   ;; (format #t "backward-trans  ~s ~%" res )
       res
))
                                                                                                 
(define (R-red-range top x rest)
   (let* ((f 0)(F 0)(c 0)(C 0)(d 0)(D 0)(n 0)(m 0))
  ;;;(format #t ":range x ~s ~%" x)
                (set! c (cadr x))
                (set! n (car c))
                (set! m (cdr c))
                (set! d (caddr x))
                (set! C *state-num*)
                 (set! f (list C))
                ;;;(set! c (cons (list 'range :ge C n) rest))
                (set! c (cons (list 'range :ge f n) rest))
                ;;;(set! d  (cons (list 'range :le C m) (concate d C)))
                (set! d  (cons (list 'range :le f m) (concate d C)))
                (if (not (eq? top :range?))
                   (set! c (list (list :or  d  c )))   ;;;;  Only here for yokubari!
                   (set! c (list (list :or  c  d )))   ;;;;  Only here for yokubari!
                )
               ;;; (format #t ":range C D  ~s ~s ~%" C D)
                (force-new-state c)   ;;;;   must be C !
               ;;; (format #t ":range C D  ~s ~s ~%" C D)
                (set! d (list (list 'range :reset f)) )
                (set! C (R-red c))
                (set! D (R-red d))
                ;;;(format #t ":range C D  ~s ~s ~%" C D)
                 (cons (get-link D) (get-link  C))
))

(define (R-red-?? top x rest)
   (let* ((exp  0))
       ;;;(set! exp   (list  (list :or  rest (append  (cadr x) rest))))
                 (if (not (eq? top :??))
                    (set! exp   (list  (list :or  (append  (cadr x) rest) rest)))
                    (set! exp   (list  (list :or  rest (append  (cadr x) rest))))
                 )
                 (R-red exp)
))


(define (R-red-not top x rest)      
   (let* ((f 0)(F 0)(c 0)(C 0)(d 0)(D 0))
     (set! C (R-red (cadr x)))
                (set! c (list (list  top (get-link C))))
                (set! C *state-num*)
                (force-new-state c)
                (set-trans C c)
                (set! D (R-red rest))
                (cons C (get-link  D))
))


(define (R-red-if top x rest)    ;;; top= :if  
   (let* ((f 0)(F 0)(c 0)(C 0)(d 0)(D 0))
      ;;;(format #t "if-reduce x ~s ~%" x)
                (set! f (cadr x))
                (when (and (pair? f) (keyword? (car f)))  ;;;  :test  Need!
                                  ;;;;  because only number is confused to Group-ref
                 ;;; (or (eq? (car f) :test) (eq? (car f) :try)))  ;;; sakiyomi
                   (set! F (R-red (cadr f)))
                   (set! f (list (car f) (get-link F)))
                )
                (set! C (R-red (caddr x)))
                (set! d  (cdddr x))
                (if (pair? d)
                      (begin
                         (set! D (R-red (car d)))
                          (set! d (get-link D))
                      )
                      (set! d 0)
                )
               (set! c
                    (list (list  top f (get-link C) d)) 
               )
                (set! C (new-state c))
                (set-trans C c)
                (set! D (R-red rest))
                ;;;;(format #t "reduce if E ~s C ~s D ~s~%" E C D)
                (cons C (get-link  D))
))

;;;;;New by stack version this is very simple!!!   
(define (R-red-atomic top x rest)
   (let* ((C 0)(D 0))
       (set! C (R-red (cadr x)))
      (set! D (R-red rest))
      (cons (get-link C) (get-link D))
))

(define (R-atomic-pat  x rest)  ;;; x is pat
   (let* ((C 0)(D 0))
       (set! C (R-red  (list x)))   ;;;; to exp
      (set! D (R-red rest))
      (cons (get-link C) (get-link D))
))



