;;; make-pdf.ss
;;;  2012.12.14
;;;   2013.12.23 verbatim error fix
;;;  2019.4.28   bibtex
;;;     file date
;;;  2020.1.15 index

(define (date)
   ;;;  mac  linux  OK
   (exec (tk-string "date +" "\" %F %T\""))
)
(define ppp-FILE #f)
(define Temp-keep #f)

(define Head-sub
"\\documentclass[10pt]{jarticle}
\\topmargin = 0cm
\\oddsidemargin = 0cm
\\textwidth = 18cm
\\textheight = 23.3cm

\\usepackage{enumerate}
\\usepackage{amsthm}
\\usepackage{multicol}
\\newtheorem{thm}{Theorem}
\\newtheorem{lem}{Lemma}

\\newcommand{\\disp}{\\displaystyle}

\\pagestyle{empty}

\\begin{document}
")
(define (make-Head)
  (let * ((font "" ))
      (if Tex-multi (set! font "\\fontsize{8pt}{14pt}\\selectfont"))
      (tk-string 
        Head-sub font #\newline
        "{\\large " "\\verb!" ppp-FILE "!" (date) "}" #\newline 
       )
))

(define Tail
"
\\end{document}
")

 
(define (make-temp-dir file)
  (let* ((y (file-sepa-ext file )) (base  (car y) )(ext (cdr y))
          (base-file base)  (x 0)(temp 0))
    (if (not(equal?  ext ""))
       (set! base-file  (tk-string base  "_" (substring ext 1 :end)))
    )
       (set! base (tk-string "Temp-" base-file))
       (set! temp (tk-string base x ))
     (while (file-exists? temp)
       (inc x)
       (set! temp (tk-string base x ))
     )
     (cons temp  base-file)  ;; ( temp-dir . base_ext))
))

(define (out-str str out)
    (format out "~a" str)
)

(define (tab->space x)
  (let* ((ind 0) (tab (string #\tab)))
     (set! ind (string-index tab x))
    (if ind 
       (string-append (substring x 0 ind) "    "
            (tab->space (substring x (+ ind 1) :end)))
        x
    )
)) 
       
(define (verbatim x)
   (let* ((h  33)     ;;; #\!
           (s 0))
       (while (string-index (string (integer->char h)) x)
           (inc h)
           (when (> h 126)  ;;;;error
               (format #t "Too complex text!!!  exit!~%~a~%" x)
               (bye)
           )
       )
      (set! h (integer->char h))
      (tk-string 
           "\\end{verbatim}" #\newline
           "\\verb" h x h #\newline
           "\\begin{verbatim}" #\newline  )
))

(define (check-verbatim x)
   (let* ((verb "\\end{verbatim}")
          (ind (string-index verb x)) )
      (if ind 
       (cons #t (verbatim x))
        (cons #f x)
     )
))

(define (count-space x)
;;;  return count of space in head of x
;;;  but may be  all  space
 (let* ((c 0) (len (string-length x)))
   (while (and (< c len) (eq? (string-ref x c) #\space))
      (inc c)
   )
 c
))
(define (check-space x)
   (let* ((c (count-space x)) (len (string-length x)))
      (if (= c len)
          ""  ;;;  all space!
          x
      )
))
      

 
(define (out-file file out)
  (let* ((in (open-input-file file)) (x (read-line in))(MAXL 85) (fx 0)(top 0)
          (indent ""))
    ;;; (if Tex-multi (set! MAXL 55))  ;;; for 10pt scriptsize
    ;;(if Tex-multi (set! MAXL 45)) ;;;  for 8pt
    (if Tex-multi (set! MAXL 50)) ;;;  for 8pt
    (while (not (eof-object? x))
       (set! x (tab->space x))  ;;; tab-> "    "
       (set! x (check-space x))
       (set! fx (check-verbatim x))
       (set! x (cdr fx))
    
         ;;;;  if verb+ ...+  , this must be single line!!!
         (when (not (car fx))  ;;; not verb
            (when (> (string-length x) MAXL)
                (set! top (count-space x))
                (if (> top 16) (set!  top 16))
                (if (< top 2) (set!  top 8))
               (set! indent (make-string top #\space))
            )
            (while (> (string-length x) MAXL)
                (format out "~a~%" (substring x 0 MAXL))
                (set! x (string-append indent (substring x MAXL :end)))
               ;;;    (set! x (substring x MAXL :end))
            )
         )
       (format out "~a~%" x)
       (set! x (read-line in))
    )
   (close-port in)
))

(define (action file)
  (let* ((y (file-sepa-ext file ))
            (base (car y)) (ext (cdr y)))
     (cond  ((equal? ".tex" ext) (tex file base))
            ((equal? ".pdf" ext) )
            ((equal? ".dvi" ext) (dvi->pdf file))
            (else  (ppp file))
     )
))

(define (tex-simple  file base)
    (system (tk-string "platex " file))
    (system (tk-string "platex " file))
     (system (tk-string "dvipdfmx  "  base ".dvi"))
)

(define (tex file base)
   (let* ((ct  0))
    (system (tk-string "platex " file))
     (inc  ct)
     (inc ct (bibtex file base))
     (set! ct (indextex file base ct))
     ;;(format #t "PLATEX ~a~%" ct)
     (system (tk-string "platex " file))
     (inc ct)
     (format #t "PLATEX ~a times~%" ct)
     (system (tk-string "dvipdfmx  "  base ".dvi"))
))

(define (bibtex file base)
   (let* ((bib (exec (tk-string "grep bibdata " base ".aux"))))
     (if  (equal? bib "")
            0
        (begin
          (system (tk-string "bibtex " base))
          (system (tk-string "platex " file))
          1
        )
    )
))

(define (indextex file base ct)
   (let* ((ind (file-exists? (tk-string base ".idx"))))
     (if  ind 
       (begin
         (when  (= ct  1) (system (tk-string "platex " file))(inc ct))
         (system (tk-string "makeindex " base))
         (system (tk-string "platex " file))
         (+ ct 1)
       )
      ct
     )
))

       
        

(define (dvi->pdf file)
     (system (tk-string "dvipdfmx  "  file))
)

(define Tex-multi #t)

(define (ppp file)
   (let* ((x 0) (temp2 (make-temp-dir file))
              (dir (car temp2))(base (cdr temp2))
              (pdf (tk-string base ".pdf")) 
              (texfile (tk-string base ".tex")) (out 0))
        (system (tk-string "mkdir " dir))
        (set! out (open-output-file (tk-string dir "/" texfile)))
              
    ;;; (system (tk-string "cat Templ-head " file " Templ-tail>" temp))
    (set!  ppp-FILE file)
    (out-str (make-Head) out)
      (when Tex-multi
         (format out "\\setlength{\\columnseprule}{0.1pt}~%")
         (format out "\\begin{multicols}{2}~%")
         ;;;(format out "\\scriptsize~%")
      )
     (format out "\\begin{verbatim}~%")
    (out-file file out)
     (format out "\\end{verbatim}~%")
      (when Tex-multi
         (format out "\\end{multicols}~%")
      )
    (out-str Tail out)
    (close-port out)
    (chdir dir)
    (tex-simple texfile base)
    (if(file-exists? pdf)
      (begin
       (system (tk-string "mv " pdf " ../"))
       (chdir "..")
        (if (not Temp-keep)
          (system (tk-string "rm -fR " dir))
        )
      ) 
      (chdir "..")
    )
   ;;(remove-file temp)
))
;;(trace ppp make-temp-dir)

 (define (help)
     (format #t "usage:  makepdf [option] file1 ...~%")
     (format #t "        *.tex   -> platex,platex->  pdf~%")
     (format #t "        *.dvi   ->   pdf~%")
     (format #t "        bibtex  is auto~%")
     ;;(format #t "    other[default] ->  temp-xxxxx.tex ,  temp-xxxxxx.pdf~%")
     (format #t "    file.other  ->  file.pdf~%")
     (format #t "  option:   effect for file.other~%")
     (format #t "         -m   two-column[default]  output~%")
     (format #t "         -s   one-column output~%")
     (format #t "         -c   [default]  use Temp-dir and clean it~%")
     (format #t "         -t    use Temp-dir and keep it~%")
     (bye)
)

(let* ((f #f))
  (when (not (pair? *argv*)) (help))
 (while (pair? *argv*)
   (set! f   (pop *argv*))
   (cond ( (equal? f "-m")
           (set! Tex-multi #t))
         ( (equal? f "-s")
           (set! Tex-multi #f))
         ( (equal? f "-t")
            (set! Temp-keep #t))
         ( (equal? f "-h")
           (help))
         (else 
             (if (file-exists? f)
                (action  f)
                (help)
             )
         )
  )
 )
)
(bye)
